/* Name: cdc162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-9
 * Tabsize: 8
 * Copyright: (c) 2007 by Koji Suzuki (suz@nmj.sakura.ne.jp)
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include "usbconfig.h"
#include "usb162.h"
#include "cdc162.h"
#include "msc162.h"


#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#define SC_VENDOR_ID	'A','T','M','E','L',' ',' ',' '
#define SC_PRODUCT_ID	'A','T','9','0','U','S','B','1','2','8',' ','M',' ','S',' ',' '
#define SC_REVISION_ID	'0','.','0','0'

#ifdef EP_MS_OUT
uint8_t msc_inquiry_data[] PROGMEM = {
	0x00, // DeviceType: Direct-access device , 0x05: CDROM
	0x80, // Medium is removable
	0x03, // Version : SPC (=SBC) (ANSI SCSI level 3 ?)
	0x02, // Response data format
	31,   // Additional Length
	0,0,0, // Reserved
	SC_VENDOR_ID,
	SC_PRODUCT_ID,
	SC_REVISION_ID,
};

uint8_t msc_maxlun;
uint32_t msc_capacity = 1024;
msc_cbw_t msc_cbw;

#define MSC_RESULT_OK		0
#define MSC_RESULT_CMD_ERR	1
#define MSC_RESULT_FRAME_ERR	2

uint8_t msc_result;

uint8_t msc_sense_key = SC_SENSE_KEY_NO_SENSE;
uint8_t msc_sense_asc = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION;

uint32u_t msc_trans_offset;
void (* msc_trans_proc)();

#define MSC_STATE_COMMAND	0
#define MSC_STATE_TRANS		1
#define MSC_STATE_STATUS	2

uint8_t msc_state = MSC_STATE_COMMAND;

static void usbmsc_putc2(uint8_t ch) {
	usbmsc_putc(0); 
	usbmsc_putc(ch); 
}

static void usbmsc_putc4(uint8_t ch) {
	usbmsc_putc2(0); 
	usbmsc_putc2(ch); 
}

void callback_msc_reset() {
	/* endpoint FIFO Reset */
	bit_set(UERST , EP_MS_IN);
	bit_clear(UERST , EP_MS_IN);
	bit_set(UERST , EP_MS_OUT);
	bit_clear(UERST , EP_MS_OUT);

	msc_state = MSC_STATE_COMMAND;
}

static void msc_trans_skip() {
	uint8_t len,i;
	if (msc_cbw.cbw.trans_len.u32 > 32) len = 32;
	else len = msc_cbw.cbw.trans_len.u32;

	if (msc_cbw.cbw.flags & 0x80) { /* Bulk IN */
		if (usbmsc_can_putc(len)) {
			for (i=0; i<len; i++) {
				usbmsc_putc(0); 
			}
			msc_cbw.cbw.trans_len.u32 -= len;
		}
	} else { /* Bulk OUT */
		if (usbmsc_can_getc(len)) {
			for (i=0; i<len; i++) {
				usbmsc_getc(); 
			}
			msc_cbw.cbw.trans_len.u32 -= len;
		}
	}
}

static void msc_trans_inquiry() {
	while ((msc_trans_offset.ub[0] < 36) && usbmsc_can_putc(1)) {
		usbmsc_putc(__LPM(msc_inquiry_data+msc_trans_offset.ub[0]));
		msc_trans_offset.ub[0]++;
		msc_cbw.cbw.trans_len.u32 --;
	}
}

static void msc_trans_request_sense() {
	if (usbmsc_can_putc(18)) {
		usbmsc_putc(0x70); // RESPONSE_CODE_SENSE
		usbmsc_putc(0x00);
		usbmsc_putc(msc_sense_key);
		usbmsc_putc4(0x00);
		usbmsc_putc(10); // ADDITIONAL_SENSE_LENGTH
		usbmsc_putc4(0x00);
		usbmsc_putc(msc_sense_asc);
		usbmsc_putc(0x00); // msc_sense_ascq
		usbmsc_putc4(0x00);
		msc_cbw.cbw.trans_len.u32 -= 18;
		msc_trans_proc = msc_trans_skip;
	}
}

#define MSC_SENSE_MODE10		(1<<0)
#define MSC_SENSE_PAGE_CODE_ERR1	(1<<1)
#define MSC_SENSE_PAGE_CODE_ERR2	(1<<2)
#define MSC_SENSE_PAGE_CODE_EXPT	(1<<3)

uint8_t msc_sense_mode;
uint8_t msc_pagecode_len;
uint8_t msc_write_protected;

static void msc_trans_mode_sense() {
	if (usbmsc_can_putc(32)) {
	    if (msc_sense_mode & MSC_SENSE_MODE10) {
		usbmsc_putc2(msc_pagecode_len); // Data length
		usbmsc_putc(0x00); // DeviceType: DEFAULT_MEDIUM_TYPE
		usbmsc_putc(msc_write_protected ? 0x80:0x00);
		usbmsc_putc4(0x00); // BLOCK_DESCRIPTOR_LENGTH
		msc_cbw.cbw.trans_len.u32 -= 8;
	    } else {
		usbmsc_putc(msc_pagecode_len); // Data length
		usbmsc_putc(0x00); // DeviceType: DEFAULT_MEDIUM_TYPE
		usbmsc_putc(msc_write_protected ? 0x80:0x00);
		usbmsc_putc(0x00); // BLOCK_DESCRIPTOR_LENGTH
		msc_cbw.cbw.trans_len.u32 -= 4;
	    }
	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_ERR1) {
		usbmsc_putc(SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x80);
		usbmsc_putc(0x03);  // READ_RETRY_COUNT
		usbmsc_putc4(0x00); // CORRECTION_SPAN,HEAD_OFFSET_COUNT
				    // ,DATA_STROBE_OFFSET,Reserved
		msc_cbw.cbw.trans_len.u32 -= 8;
	    }
	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_ERR2) {
		usbmsc_putc(0x80);  // WRITE_RETRY_COUNT
		usbmsc_putc(0x00);
		usbmsc_putc2(0x00); // RECOVERY_LIMIT
		msc_cbw.cbw.trans_len.u32 -= 4;
	    }
	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_EXPT) {
		// Page Code: Informational exceptions control page
		// See chapter 8.3.8 on SPC-2 specification
		usbmsc_putc(SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x00);
		usbmsc_putc(0x05);  // MRIE
		usbmsc_putc4(0x00); // Interval Timer 
		usbmsc_putc4(0x01); // Report Count
		msc_cbw.cbw.trans_len.u32 -= 12;
	    }
	    msc_trans_proc = msc_trans_skip;
	}
}

static void msc_trans_read_capacity() {
	if (usbmsc_can_putc(8)) {
		usbmsc_putc((uint8_t)(msc_capacity>>24));
		usbmsc_putc((uint8_t)(msc_capacity>>16));
		usbmsc_putc((uint8_t)(msc_capacity>>8));
		usbmsc_putc((uint8_t)(msc_capacity));
		usbmsc_putc(0x00); // SECTOR SIZE: 512B
		usbmsc_putc(0x00);
		usbmsc_putc(0x01);
		usbmsc_putc(0x00);
		msc_cbw.cbw.trans_len.u32 -= 8;
		msc_trans_proc = msc_trans_skip;
	}
}

static void msc_trans_read() {
}

static void msc_trans_write() {
}
#endif

void msc_poll() {
#ifdef EP_MS_OUT
	uint8_t i;
	uint8_t sc_cmd;
	uint8_t sc_len = 15;
	uint8_t pagecode;
	switch (msc_state) {
	case MSC_STATE_COMMAND:
		if (usbmsc_can_getc(31)) {
			for (i=0; i<15; i++) {
				msc_cbw.ub[i] = usbmsc_getc();
			}
			msc_trans_proc = msc_trans_skip;
			msc_trans_offset.u32 = 0;
			sc_cmd = usbmsc_getc();
			if (sc_cmd != SC_REQUEST_SENSE) {
				msc_sense_key = SC_SENSE_KEY_NO_SENSE;
				msc_sense_asc = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION;
			}
			switch (sc_cmd) {
			case SC_INQUIRY:
				msc_trans_proc = msc_trans_inquiry;
				break;
			case SC_MODE_SENSE_6:
				//sc_len -= xx;
				msc_trans_proc = msc_trans_mode_sense;
				msc_pagecode_len = 0;
				break;
#if 0
			case SC_MODE_SENSE_10:
				msc_sense_mode = MSC_SENSE_MODE10;
				msc_trans_proc = msc_trans_mode_sense;
				switch ( pagecode & SC_MSK_PAGE_CODE ) {
				case SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS: 
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT;
					msc_pagecode_len = 12 + 3;
					break;
				case SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY:
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR1;
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR2;
					msc_pagecode_len = 12 + 3;
					break;
				case SC_PAGE_CODE_ALL:
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT;
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR1;
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR2;
					msc_pagecode_len = 12 + 12 + 3;
					break;
				default:
					msc_sense_key = SC_SENSE_KEY_ILLEGAL_REQUEST;
					msc_sense_asc = SC_ASC_INVALID_FIELD_IN_CDB;
					msc_trans_proc = msc_trans_skip;
				}
				break;
#endif
			case SC_PREVENT_ALLOW_MEDIUM_REMOVAL:
				break;
			case SC_READ_10:
				msc_trans_proc = msc_trans_skip;
				break;
			case SC_READ_CAPACITY:
				msc_trans_proc = msc_trans_read_capacity;
				break;
			case SC_REQUEST_SENSE:
				msc_trans_proc = msc_trans_request_sense;
				break;
			case SC_TEST_UNIT_READY:
				break;
			case SC_VERIFY_10:
				msc_trans_proc = msc_trans_skip;
				break;
			case SC_WRITE_10:
				msc_trans_proc = msc_trans_skip;
				break;
			default:
				msc_sense_key = SC_SENSE_KEY_ILLEGAL_REQUEST;
				msc_sense_asc = SC_ASC_INVALID_COMMAND_OPERATION_CODE;
			}
			for (i=0; i<sc_len; i++) {
				usbmsc_getc();
			}
			if (msc_cbw.cbw.trans_len.u32 == 0) {
				msc_state = MSC_STATE_TRANS;
			} else {
				msc_state = MSC_STATE_STATUS;
			}
		}
		break;
	case MSC_STATE_TRANS:
		(msc_trans_proc)();
		if (msc_cbw.cbw.trans_len.u32 == 0) {
			msc_state = MSC_STATE_STATUS;
		}
		break;
	case MSC_STATE_STATUS:
		if (usbmsc_can_putc(13)) {
			msc_cbw.csw.magic[3] = 'S';
			msc_cbw.csw.status = 0;
			for (i=0; i<13; i++) {
				usbmsc_putc(msc_cbw.ub[i]); 
			}
			msc_state = MSC_STATE_COMMAND;
		}
		break;
	default:
		break;
	}
#endif
}
